import re

from .text import compact_whitespace


UPPER_LOWER_PATTERN = re.compile(r'(.)([A-Z][a-z]+)')
LOWER_UPPER_PATTERN = re.compile(r'([a-z0-9])([A-Z])')
LETTER_DIGIT_PATTERN = re.compile(r'([a-z])([0-9])')
DIGIT_LETTER_PATTERN = re.compile(r'([0-9])([a-z])')


def duplicate_selected_column_names(selected_column_names, column_names):
    suffix = '*'
    while set(column_names).intersection(
            x + suffix for x in selected_column_names):
        suffix += '*'
    return list(column_names) + [x + suffix for x in selected_column_names]


def normalize_key(
        x, word_separator=' ', separate_camel_case=False,
        separate_letter_digit=False):
    """
    Normalize key using a variation of the method described in
    http://stackoverflow.com/a/1176023/192092

    ONETwo   one two
    OneTwo   one two
    one-two  one two
    one_two  one two
    one2     one 2
    1two     1 two
    """
    if separate_camel_case:
        x = UPPER_LOWER_PATTERN.sub(r'\1 \2', x)
        x = LOWER_UPPER_PATTERN.sub(r'\1 \2', x)
    x = x.lower()
    if separate_letter_digit:
        x = LETTER_DIGIT_PATTERN.sub(r'\1 \2', x)
        x = DIGIT_LETTER_PATTERN.sub(r'\1 \2', x)
    word_separators = ['-', '_', ' ']
    if word_separator not in word_separators:
        word_separators.append(word_separator)
    word_separator_expression = '[' + ''.join(word_separators) + ']'
    word_separator_pattern = re.compile(word_separator_expression)
    x = word_separator_pattern.sub(' ', x)
    x = compact_whitespace(x)
    return x.replace(' ', word_separator)



from copy import copy


def transform_geometries(geometries, f):
    transformed_geometries = []
    for geometry in geometries:
        if hasattr(geometry, 'geoms'):
            transformed_geometry = geometry.__class__(transform_geometries(
                geometry.geoms, f))
        else:
            transformed_geometry = copy(geometry)
            transformed_geometry.coords = [f(xyz) for xyz in geometry.coords]
        transformed_geometries.append(transformed_geometry)
    return transformed_geometries


def flip_xy(xyz):
    'Flip x and y coordinates whether or not there is a z-coordinate'
    xyz = list(xyz)  # Preserve original
    xyz[0], xyz[1] = xyz[1], xyz[0]
    return tuple(xyz)


def drop_z(xyz):
    return xyz[:2]



import datetime


def get_timestamp(x=None, with_microsecond=False):
    if x is None:
        x = datetime.datetime.now()
    if with_microsecond:
        return x.strftime('%Y%m%d-%H%M-%f')
    else:
        return x.strftime('%Y%m%d-%H%M')



import random
import string


try:
    ALPHABET = string.digits + string.letters
except AttributeError:
    ALPHABET = string.digits + string.ascii_letters
RANDOM = random.SystemRandom()


def make_random_string(
        length, with_punctuation=False, with_spaces=False, alphabet=ALPHABET):
    if with_punctuation:
        alphabet += string.punctuation
    if with_spaces:
        alphabet += ' '
    return ''.join(RANDOM.choice(alphabet) for x in range(length))



import attr
import datetime
import re
import shlex
import simplejson as json
from six import string_types
from six.moves.urllib.parse import urlencode as format_query
from subprocess import CalledProcessError, Popen, check_output, PIPE, STDOUT

from .calculator import get_int
from .exceptions import InvisibleRoadsError


@attr.s
class Callback(object):

    id = attr.ib()
    datetime = attr.ib()


def run_command(command, exception_by_error=None):
    if not isinstance(command, string_types):
        command = ' '.join(command)
    command = command.split(';', 1)[0]
    return run_raw_command(command, exception_by_error)


def run_raw_command(command, exception_by_error=None):
    if isinstance(command, string_types):
        command = shlex.split(command)
    try:
        output = check_output(command, stderr=STDOUT)
    except CalledProcessError as e:
        o = e.output
        for error_text, exception in (exception_by_error or {}).items():
            if error_text in o:
                raise exception
        else:
            raise InvisibleRoadsError(o)
    except OSError as e:
        raise InvisibleRoadsError(e.strerror)
    return output.strip()


def schedule_curl_callback(
        minute_count, base_url, value_by_key=None, headers=None, method='GET',
        with_retry=True):
    # Prepare more_lines
    full_url = base_url
    more_lines = []
    if value_by_key:
        if method == 'POST':
            headers['Content-Type'] = 'application/json'
            more_lines.append('-d \'%s\'' % json.dumps(value_by_key))
        else:
            full_url += format_query(value_by_key)
    if with_retry:
        more_lines.extend(['--retry 7'])
    more_lines.extend(['-X %s' % method, full_url])
    # Prepare header_lines
    header_lines = []
    for k, v in (headers or {}).items():
        header_lines.append('-H "%s: %s"' % (k, v))
    # Schedule callback
    shell_parts = ['curl'] + header_lines + more_lines
    shell_text = ' '.join(shell_parts)
    return schedule_shell_callback(minute_count, shell_text)


def schedule_shell_callback(minute_count, shell_text):
    process = Popen([
        'at', 'now + %s minutes' % get_int(minute_count),
    ], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate((shell_text + '\n').encode('utf-8'))
    try:
        callback_id, when_string = re.search(
            'job (\d+) at (.+)', stderr.decode('utf-8')).groups()
    except AttributeError:
        raise InvisibleRoadsError('cannot schedule callback')
    return Callback(id=int(callback_id), datetime=datetime.datetime.strptime(
        when_string, '%a %b %d %H:%M:%S %Y'))


def cancel_shell_callback(callback_id):
    process = Popen(['atrm', str(callback_id)])
    process.wait()


def format_variables_as_shell_script(d):
    lines = []
    for k, v in d.items():
        if v is None:
            v = ''
        lines.append('%s=%s' % (k.upper(), v))
    return '\n'.join(lines)



import attr
import codecs
import configparser
import functools
import re
import shlex
from argparse import ArgumentError, ArgumentParser
from collections import OrderedDict
from importlib import import_module
from os.path import dirname, expanduser, isabs, join, relpath
from six import string_types

from .calculator import get_int
from .disk import expand_path, resolve_relative_path
from .iterable import merge_dictionaries
from .log import format_summary, get_log


L = get_log(__name__)
SECTION_TEMPLATE = '[%s]\n%s\n'


class Settings(dict):

    def set(self, settings, prefix, key, default=None, parse=None):
        value = set_default(settings, prefix + key, default, parse)
        self[key] = value
        return value


class StoicArgumentParser(ArgumentParser):

    def add_argument(self, *args, **kw):
        try:
            return super(StoicArgumentParser, self).add_argument(*args, **kw)
        except ArgumentError:
            pass


class TerseArgumentParser(StoicArgumentParser):

    def add(self, argument_name, *args, **kw):
        d = {}
        if argument_name.endswith('_path'):
            d['metavar'] = 'PATH'
        elif '_as_percent_of_' in argument_name:
            d['metavar'] = 'FLOAT'
            d['type'] = float
        elif argument_name.endswith('_year'):
            d['metavar'] = 'YEAR'
            d['type'] = int
        elif argument_name.endswith('_in_years'):
            d['metavar'] = 'INTEGER'
            d['type'] = int
        d.update(kw)
        self.add_argument('--' + argument_name, *args, **d)


class RawCaseSensitiveConfigParser(configparser.RawConfigParser):
    optionxform = str


def set_default(settings, key, default=None, parse=None):
    value = settings.get(key, default)
    if key not in settings:
        L.warn('using default %s = %s' % (key, value))
    elif value in ('', None):
        L.warn('missing %s' % key)
    elif parse:
        value = parse(value)
    settings[key] = value
    return value


def save_settings(
        configuration_path, settings_by_section_name,
        suffix_format_packs=None):
    configuration_file = codecs.open(configuration_path, 'w', encoding='utf-8')
    configuration_file.write(format_settings(
        settings_by_section_name, suffix_format_packs) + '\n')
    return configuration_path


def load_relative_settings(configuration_path, section_name):
    settings = OrderedDict()
    configuration_folder = dirname(expand_path(configuration_path))
    for k, v in load_settings(configuration_path, section_name).items():
        if k.endswith('_path') or k.endswith('_folder'):
            v = resolve_relative_path(expanduser(v), configuration_folder)
        settings[k] = v
    return settings


def load_settings(configuration_path, section_name):
    configuration = RawCaseSensitiveConfigParser()
    configuration.read(configuration_path, 'utf-8')
    try:
        items = configuration.items(section_name)
    except configparser.NoSectionError:
        items = []
    return OrderedDict(items)


def format_settings(settings_by_section_name, suffix_format_packs=None):
    configuration_parts = []
    for section_name, settings in settings_by_section_name.items():
        configuration_parts.append(SECTION_TEMPLATE % (
            section_name, format_summary(settings, suffix_format_packs)))
    return '\n'.join(configuration_parts).strip()


def gather_settings(settings, prefix, parse_setting=None):
    d = {}
    prefix_pattern = re.compile('^' + prefix.replace('.', r'\.'))
    if not parse_setting:
        parse_setting = parse_raw_setting
    for k, v in settings.items():
        if not k.startswith(prefix):
            continue
        d = merge_dictionaries(d, parse_setting(prefix_pattern.sub('', k), v))
    return d


def parse_raw_setting(k, v):
    return {k: v}


def resolve_attribute(attribute_spec):
    # Modified from pkg_resources.EntryPoint.resolve()
    if not attribute_spec or not hasattr(attribute_spec, 'split'):
        return attribute_spec
    module_url, attributes_string = attribute_spec.split(':')
    module = import_module(module_url)
    try:
        attribute = functools.reduce(
            getattr, attributes_string.split('.'), module)
    except AttributeError:
        raise ImportError('could not resolve attribute (%s)' % module_url)
    return attribute


def split_arguments(command_string):
    try:
        xs = shlex.split(command_string)
    except UnicodeEncodeError:
        xs = shlex.split(command_string.encode('utf-8'))
    return [x.strip() for x in xs]


def make_absolute_paths(d, folder):
    d = OrderedDict(d)
    for k, v in d.items():
        if hasattr(v, 'items'):
            v = make_absolute_paths(v, folder)
        elif is_path_key(k) and v:
            v = expanduser(v)
            if not isabs(v):
                v = join(folder, v)
        d[k] = v
    return d


def make_relative_paths(d, folder, with_force=True):
    d = OrderedDict(d)
    for k, v in d.items():
        if hasattr(v, 'items'):
            v = make_relative_paths(v, folder, with_force)
        elif is_path_key(k):
            v = make_relative_path_safely(v, folder, with_force)
        d[k] = v
    return d


def make_relative_path_safely(path, folder, with_force=True):
    expanded_path = expanduser(path)
    if not isabs(expanded_path):
        return expanded_path
    expanded_folder = expanduser(folder)
    relative_path = relpath(expanded_path, expanded_folder)
    if relative_path.startswith('..'):
        if with_force:
            return ''
        else:
            return expanded_path
    return relative_path


def is_path_key(k):
    return k.endswith('_path') or k.endswith('_folder')


def encode_object(o):
    """Use with json.dumps to serialize classes to JSON.

    x = json.dumps(d, default=encode_object)
    """
    if hasattr(o, '__attrs_attrs__'):
        d = attr.asdict(o)
        d['__class__'] = o.__class__.__name__
        return d
    raise TypeError(repr(o) + ' is not JSON serializable')


def define_decode_object(class_by_name):
    """Use with json.loads to deserialize classes from JSON.

    decode_object = define_decode_object(globals())
    d = json.loads(x, object_hook=decode_object)
    """
    def decode_object(d):
        if '__class__' in d:
            class_name = d.pop('__class__')
            Class = class_by_name[class_name]
            return Class(**d)
        return d
    return decode_object


def define_gather_numbers(expression):

    def gather_numbers(settings):
        numbers = []
        number_pattern = re.compile(expression)
        for k, v in settings.items():
            match = number_pattern.match(k)
            if not match:
                continue
            numbers.append(int(match.group(1)))
        return sorted(numbers)

    return gather_numbers


def parse_integers(x):
    return [int(y) for y in parse_list(x)]


def parse_list(x):
    if isinstance(x, string_types):
        x = x.split()
    return x


def parse_minute_count(x):
    return get_int(parse_second_count(x) / 60.)


def parse_second_count(x):
    if isinstance(x, int):
        return x
    if not isinstance(x, string_types):
        raise ValueError
    try:
        x_count, x_unit = re.match('(\d+)([hms])', x.strip()).groups()
    except AttributeError:
        raise ValueError
    x_count = int(x_count)
    if x_unit == 'h':
        x_count *= 60 * 60
    elif x_unit == 'm':
        x_count *= 60
    return x_count



import re
import sys


WHITESPACE_PATTERN = re.compile(r'\s+', re.MULTILINE)
PUNCTUATION_PATTERN = re.compile(r'[^a-zA-Z\s]+')


def compact_whitespace(string):
    return WHITESPACE_PATTERN.sub(' ', string).strip()


def cut_and_strip(x, separator):
    head, separator, tail = x.partition(separator)
    return head.strip(), tail.strip()


def has_whitespace(string):
    return WHITESPACE_PATTERN.search(string) is not None


def parse_words(x):
    return x.replace(',', ' ').split()


def remove_punctuation(string):
    return compact_whitespace(PUNCTUATION_PATTERN.sub(' ', string))


def strip_whitespace(string):
    return WHITESPACE_PATTERN.sub('', string)


def unicode_safely(x):
    # http://stackoverflow.com/a/23085282/192092
    try:
        return x.decode(sys.getfilesystemencoding())
    except (AttributeError, UnicodeEncodeError):
        return x



from __future__ import print_function

try:
    from logging import NullHandler
except ImportError:  # Python 2.6
    from logging import Handler

    class NullHandler(Handler):

        def emit(self, record):
            pass
finally:
    from logging import getLogger

import re
import traceback
from collections import OrderedDict
from os.path import expanduser
from six import string_types
from sys import stderr

from .disk import COMMAND_LINE_HOME


INDENT = ' ' * 2


def get_log(name):
    log = getLogger(name)
    log.addHandler(NullHandler())
    return log


def filter_nested_dictionary(value_by_key, f):
    d = OrderedDict()
    for k, v in value_by_key.items():
        if f(k):
            continue
        if hasattr(v, 'items'):
            v = filter_nested_dictionary(v, f)
        d[k] = v
    return d


def print_error(x, *args):
    'Print to standard error stream'
    if not isinstance(x, string_types):
        x = str(x)
    print(x % args, file=stderr)


def stylize_dictionary(value_by_key, suffix_format_packs):
    d = {}
    for key, value in value_by_key.items():
        for suffix, format_value in suffix_format_packs:
            if key.endswith(suffix):
                value = format_value(value)
                break
        d[key] = value
    return d


def format_summary(value_by_key, suffix_format_packs=None):
    suffix_format_packs = list(suffix_format_packs or [])
    suffix_format_packs.extend([
        ('_folder', format_path),
        ('_path', format_path),
    ])
    return format_nested_dictionary(OrderedDict(
        value_by_key), suffix_format_packs)


def format_nested_dictionary(
        value_by_key, suffix_format_packs=None, prefix=''):
    parts = []
    for key, value in value_by_key.items():
        left_hand_side = prefix + str(key)
        if hasattr(value, 'items'):
            parts.append(format_nested_dictionary(
                value, suffix_format_packs, left_hand_side + '.'))
            continue
        for suffix, format_value in suffix_format_packs or []:
            if key.endswith(suffix):
                parts.append(format_assignment(
                    left_hand_side, format_value(value)))
                break
        else:
            if not isinstance(value, string_types):
                value = str(value)
            if '\n' in value:
                value = format_indented_block(value)
            parts.append(format_assignment(
                left_hand_side, value))
    return '\n'.join(parts)


def format_assignment(left_hand_side, right_hand_side):
    left_hand_side = left_hand_side.strip()
    if right_hand_side.startswith('\n'):
        operator = ' ='
    else:
        operator = ' = '
        right_hand_side = right_hand_side.strip()
    return left_hand_side + operator + right_hand_side


def format_path(x):
    return re.sub(r'^' + expanduser('~'), COMMAND_LINE_HOME, x)


def format_indented_block(x, indent=INDENT):
    return '\n' + '\n'.join(indent + line for line in x.splitlines())


def format_decimal(x, fractional_digit_count=2):
    template = '{:,.%sf}' % fractional_digit_count
    return template.format(x)


def format_delta(x):
    template = '{:+,}'
    return template.format(x)


def format_number(x):
    return format_decimal(x, fractional_digit_count=0)


def get_nested_dictionary(nested_lists):
    d = OrderedDict()
    for k, v in nested_lists:
        if k.endswith('_'):
            try:
                v = get_nested_dictionary(v)
            except TypeError:
                pass
            else:
                k = k[:-1]
        d[k] = v
    return d


def get_nested_lists(nested_dictionary):
    xs = []
    for k, v in nested_dictionary.items():
        if hasattr(v, 'items'):
            v = get_nested_lists(v)
            k = k + '_'
        xs.append((k, v))
    return xs


def parse_nested_dictionary_from(raw_dictionary, max_depth=float('inf')):
    value_by_key = OrderedDict()
    for key, value in OrderedDict(raw_dictionary).items():
        key_parts = key.split('.')
        d = value_by_key
        depth = 0
        while key_parts:
            key_part = key_parts.pop(0)
            if len(key_parts) and depth < max_depth:
                if key_part not in d:
                    d[key_part] = OrderedDict()
                d = d[key_part]
                depth += 1
            else:
                d['.'.join([key_part] + key_parts)] = value
                break
    return value_by_key


def parse_nested_dictionary(text, is_key=lambda x: True, indent=INDENT):
    raw_dictionary, key = OrderedDict(), None
    for line in text.splitlines():
        if line.startswith(indent):
            if key is not None:
                value = line[2:].rstrip()
                raw_dictionary[key].append(value)
            continue
        try:
            key, value = line.split(' = ', 1)
            if not is_key(key):
                key = None
        except ValueError:
            key = None
        if not key:
            continue
        key = key.strip()
        value = value.strip()
        raw_dictionary[key] = [value]
    d = OrderedDict()
    for k, v in raw_dictionary.items():
        this_dictionary = d
        for key in k.split('.'):
            key = key.strip()
            last_dictionary = this_dictionary
            try:
                this_dictionary = this_dictionary[key]
            except KeyError:
                this_dictionary[key] = {}
                this_dictionary = this_dictionary[key]
        last_dictionary[key] = '\n'.join(v).strip()
    return d


def log_traceback(log, d=''):
    log.error(traceback.format_exc() + str(d))
